"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class SaberToothedTiger(Monster):
    """
    Keen Smell.
      The tiger has advantage on Wisdom (Perception) checks that rely on
      smell.
    Pounce.
      If the tiger moves at least 20 ft. straight toward a creature and then
      hits it with a claw attack on the same turn, that target must succeed
      on a DC 14 Strength saving throw or be knocked prone. If the target is
      prone, the tiger can make one bite attack against it as a bonus
      action.
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 10 (1d10
      + 5) piercing damage.
    Claw.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 12 (2d6
      + 5) slashing damage.
    """
    name = "Saber-Toothed Tiger"
    description = "Large beast, unaligned"
    challenge_rating = 2
    armor_class = 12
    skills = "Perception +3, Stealth +6"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(18)
    dexterity = Ability(14)
    constitution = Ability(15)
    intelligence = Ability(3)
    wisdom = Ability(12)
    charisma = Ability(8)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 52
    hit_dice = "7d10"
    spells = []


class Sahuagin(Monster):
    """
    Blood Frenzy.
      The sahuagin has advantage on melee attack rolls against any creature
      that doesn't have all its hit points.
    Limited Amphibiousness.
      The sahuagin can breathe air and water, but it needs to be submerged
      at least once every 4 hours to avoid suffocating.
    Shark Telepathy.
      The sahuagin can magically command any shark within 120 feet of it,
      using a limited telepathy.
    Multiattack.
      The sahuagin makes two melee attacks: one with its bite and one with
      its claws or spear.
    Bite.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 3 (1d4 +
      1) piercing damage.
    Claws.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 3 (1d4 +
      1) slashing damage.
    Spear.
      Melee or Ranged Weapon Attack: +3 to hit, reach 5 ft. or range 20/60
      ft., one target. Hit: 4 (1d6 + 1) piercing damage, or 5 (1d8 + 1)
      piercing damage if used with two hands to make a melee attack.
    """
    name = "Sahuagin"
    description = "Medium humanoid, lawful evil"
    challenge_rating = 0.5
    armor_class = 12
    skills = "Perception +5"
    senses = "Darkvision 120 ft., Passive Perception 15"
    languages = "Sahuagin"
    strength = Ability(13)
    dexterity = Ability(11)
    constitution = Ability(12)
    intelligence = Ability(12)
    wisdom = Ability(13)
    charisma = Ability(9)
    speed = 30
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 22
    hit_dice = "4d8"
    spells = []


class Salamander(Monster):
    """
    Heated Body.
      A creature that touches the salamander or hits it with a melee attack
      while within 5 ft. of it takes 7 (2d6) fire damage.
    Heated Weapons.
      Any metal melee weapon the salamander wields deals an extra 3 (1d6)
      fire damage on a hit (included in the attack).
    Multiattack.
      The salamander makes two attacks: one with its spear and one with its
      tail.
    Spear.
      Melee or Ranged Weapon Attack: +7 to hit, reach 5 ft. or range 20
      ft./60 ft., one target. Hit: 11 (2d6 + 4) piercing damage, or 13 (2d8
      + 4) piercing damage if used with two hands to make a melee attack,
      plus 3 (1d6) fire damage.
    Tail.
      Melee Weapon Attack: +7 to hit, reach 10 ft., one target. Hit: 11 (2d6
      + 4) bludgeoning damage plus 7 (2d6) fire damage, and the target is
      grappled (escape DC 14). Until this grapple ends, the target is
      restrained, the salamander can automatically hit the target with its
      tail, and the salamander can't make tail attacks against other
      targets.
    """
    name = "Salamander"
    description = "Large elemental, neutral evil"
    challenge_rating = 5
    armor_class = 15
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Ignan"
    strength = Ability(18)
    dexterity = Ability(14)
    constitution = Ability(15)
    intelligence = Ability(11)
    wisdom = Ability(10)
    charisma = Ability(12)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 90
    hit_dice = "12d10"
    spells = []


class Satyr(Monster):
    """
    Magic Resistance.
      The satyr has advantage on saving throws against spells and other
      magical effects.
    Ram.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 6 (2d4 +
      1) bludgeoning damage.
    Shortsword.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 6 (1 d6
      + 3) piercing damage.
    Shortbow.
      Ranged Weapon Attack: +5 to hit, range 80/320 ft., one target. Hit: 6
      (1d6 + 3) piercing damage.
    Variant: Panpipes.
      Gentle Lullaby. The creature falls asleep and is unconscious for 1
      minute. The effect ends if the creature takes damage or if someone
      takes an action to shake the creature awake.
    """
    name = "Satyr"
    description = "Medium fey, chaotic neutral"
    challenge_rating = 0.5
    armor_class = 14
    skills = "Perception +2, Performance +6, Stealth +5"
    senses = "Passive Perception 12"
    languages = "Common, Elvish, Sylvan"
    strength = Ability(12)
    dexterity = Ability(16)
    constitution = Ability(11)
    intelligence = Ability(12)
    wisdom = Ability(10)
    charisma = Ability(14)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 31
    hit_dice = "7d8"
    spells = []


class Scorpion(Monster):
    """
    Sting.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one creature. Hit: 1
      piercing damage, and the target must make a DC 9 Constitution saving
      throw, taking 4 (1d8) poison damage on a failed save, or half as much
      damage on a successful one.
    """
    name = "Scorpion"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 11
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 9"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(11)
    constitution = Ability(8)
    intelligence = Ability(1)
    wisdom = Ability(8)
    charisma = Ability(2)
    speed = 10
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 1
    hit_dice = "1d4"
    spells = []


class Scout(Monster):
    """
    Keen Hearing and Sight.
      The scout has advantage on Wisdom (Perception) checks that rely on
      hearing or sight.
    Multiattack.
      The scout makes two melee attacks or two ranged attacks.
    Shortsword.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 5 (1d6 +
      2) piercing damage.
    Longbow.
      Ranged Weapon Attack: +4 to hit, ranged 150/600 ft., one target. Hit:
      6 (1d8 + 2) piercing damage.
    """
    name = "Scout"
    description = "Medium humanoid, any alignment"
    challenge_rating = 0.5
    armor_class = 13
    skills = "Nature +4, Perception +5, Stealth +6, Survival +5"
    senses = "Passive Perception 15"
    languages = "any one language (usually Common)"
    strength = Ability(11)
    dexterity = Ability(14)
    constitution = Ability(12)
    intelligence = Ability(11)
    wisdom = Ability(13)
    charisma = Ability(11)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 16
    hit_dice = "3d8"
    spells = []


class SeaHag(Monster):
    """
    Amphibious.
      The hag can breathe air and water.
    Horrific Appearance.
      Any humanoid that starts its turn within 30 feet of the hag and can
      see the hag's true form must make a DC 11 Wisdom saving throw. On a
      failed save, the creature is frightened for 1 minute. A creature can
      repeat the saving throw at the end of each of its turns, with
      disadvantage if the hag is within line of sight, ending the effect on
      itself on a success. If a creature's saving throw is successful or the
      effect ends for it, the creature is immune to the hag's Horrific
      Appearance for the next 24 hours.
      
      Unless the target is surprised or the revelation of the hag's true
      form is sudden, the target can avert its eyes and avoid making the
      initial saving throw. Until the start of its next turn, a creature
      that averts its eyes has disadvantage on attack rolls against the hag.
    Claws.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 10 (2d6
      + 3) slashing damage.
    Death Glare.
      The hag targets one frightened creature she can see within 30 ft. of
      her. If the target can see the hag, it must succeed on a DC 11 Wisdom
      saving throw against this magic or drop to 0 hit points.
    Illusory Appearance.
      The hag covers herself and anything she is wearing or carrying with a
      magical illusion that makes her look like an ugly creature of her
      general size and humanoid shape. The effect ends if the hag takes a
      bonus action to end it or if she dies.
      
      The changes wrought by this effect fail to hold up to physical
      inspection. For example, the hag could appear to have no claws, but
      someone touching her hand might feel the claws. Otherwise, a creature
      must take an action to visually inspect the illusion and succeed on a
      DC 16 Intelligence (Investigation) check to discern that the hag is
      disguised.
    """
    name = "Sea Hag"
    description = "Medium fey, chaotic evil"
    challenge_rating = 2
    armor_class = 14
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 11"
    languages = "Aquan, Common, Giant"
    strength = Ability(16)
    dexterity = Ability(13)
    constitution = Ability(16)
    intelligence = Ability(12)
    wisdom = Ability(12)
    charisma = Ability(13)
    speed = 30
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 52
    hit_dice = "7d8"
    spells = []


class SeaHorse(Monster):
    """
    Water Breathing.
      The sea horse can breathe only underwater.
    """
    name = "Sea Horse"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 11
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(1)
    dexterity = Ability(12)
    constitution = Ability(8)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(2)
    speed = 0
    swim_speed = 20
    fly_speed = 0
    climb_speed = 0
    hp_max = 1
    hit_dice = "1d4"
    spells = []


class Shadow(Monster):
    """
    Amorphous.
      The shadow can move through a space as narrow as 1 inch wide without
      squeezing.
    Shadow Stealth.
      While in dim light or darkness, the shadow can take the Hide action as
      a bonus action. Its stealth bonus is also improved to +6.
    Sunlight Weakness.
      While in sunlight, the shadow has disadvantage on attack rolls,
      ability checks, and saving throws.
    Strength Drain.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 9 (2d6
      + 2) necrotic damage, and the target's Strength score is reduced by
      1d4. The target dies if this reduces its Strength to 0. Otherwise, the
      reduction lasts until the target finishes a short or long rest.
      
      If a non-evil humanoid dies from this attack, a new shadow rises from
      the corpse 1d4 hours later.
    """
    name = "Shadow"
    description = "Medium undead, chaotic evil"
    challenge_rating = 0.5
    armor_class = 12
    skills = "Stealth +4"
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = ""
    strength = Ability(6)
    dexterity = Ability(14)
    constitution = Ability(13)
    intelligence = Ability(6)
    wisdom = Ability(10)
    charisma = Ability(8)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 16
    hit_dice = "3d8"
    spells = []


class ShadowDemon(Monster):
    """Incorporeal Movement.
      The demon can move through other creatures and objects as if
      they were difficult terrain. It takes 5 (1d10) force damage if
      it ends its turn inside an object.
    Light Sensitivity.
      While in bright light, the demon has disadvantage on attack
      rolls, as well as on Wisdom (Perception) checks that rely on
      sight.
    Shadow Stealth.
      While in dim light or darkness, the demon can take the Hide
      action as a bonus action.
    Claws.
      *Melee Weapon Attack:* +5 to hit, reach 5ft., one
      creature. *Hit:* 10 (2d6+3) psychic damage or, if the demon
      had advantage on the attack roll, 17 (4d6+3) psychic damage.
    """

    name = "Shadow Demon"
    description = "Medium fiend (demon), chaotic evil"
    challenge_rating = 4
    armor_class = 13
    skills = "Stealth +7"
    saving_throws = "Dex +5, Cha +4"
    damage_vulnerabilities = "radiant"
    damage_resistances = (
        "acid, fire, necrotic, thunder; bludgeoning, piercing, and slashing from"
        " nonmagical attacks"
    )
    damage_immunities = "cold, lightning, poison"
    condition_immunities = (
        "exhaustion, grappled, paralyzed, petrified, poisoned, prone, restrained"
    )
    senses = "Darkvision 120 ft., Passive Perception 11"
    languages = "Abyssal, telepathy 120 ft."
    strength = Ability(1)
    dexterity = Ability(17)
    constitution = Ability(12)
    intelligence = Ability(14)
    wisdom = Ability(13)
    charisma = Ability(14)
    speed = 30
    fly_speed = 30
    hp_max = 66
    hit_dice = "12d8 + 12"


class ShamblingMound(Monster):
    """
    Lightning Absorption.
      Whenever the shambling mound is subjected to lightning damage, it
      takes no damage and regains a number of hit points equal to the
      lightning damage dealt.
    Multiattack.
      The shambling mound makes two slam attacks. If both attacks hit a
      Medium or smaller target, the target is grappled (escape DC 14), and
      the shambling mound uses its Engulf on it.
    Slam.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 13 (2d8
      + 4) bludgeoning damage.
    Engulf.
      The shambling mound engulfs a Medium or smaller creature grappled by
      it. The engulfed target is blinded, restrained, and unable to breathe,
      and it must succeed on a DC 14 Constitution saving throw at the start
      of each of the mound's turns or take 13 (2d8 + 4) bludgeoning damage.
      If the mound moves, the engulfed target moves with it. The mound can
      have only one creature engulfed at a time.
    """
    name = "Shambling Mound"
    description = "Large plant, unaligned"
    challenge_rating = 5
    armor_class = 15
    skills = "Stealth +2"
    senses = "Blindsight 60 ft. (blind beyond this radius), Passive Perception 10"
    languages = ""
    strength = Ability(18)
    dexterity = Ability(8)
    constitution = Ability(16)
    intelligence = Ability(5)
    wisdom = Ability(10)
    charisma = Ability(5)
    speed = 20
    swim_speed = 20
    fly_speed = 0
    climb_speed = 0
    hp_max = 136
    hit_dice = "16d10"
    spells = []


class ShieldGuardian(Monster):
    """
    Bound.
      The shield guardian is magically bound to an amulet. As long as the
      guardian and its amulet are on the same plane of existence, the
      amulet's wearer can telepathically call the guardian to travel to it,
      and the guardian knows the distance and direction to the amulet. If
      the guardian is within 60 feet of the amulet's wearer, half of any
      damage the wearer takes (rounded up) is transferred to the guardian.
    Regeneration.
      The shield guardian regains 10 hit points at the start of its turn if
      it has at least 1 hit. point.
    Spell Storing.
      A spellcaster who wears the shield guardian's amulet can cause the
      guardian to store one spell of 4th level or lower. To do so, the
      wearer must cast the spell on the guardian. The spell has no effect
      but is stored within the guardian. When commanded to do so by the
      wearer or when a situation arises that was predefined by the
      spellcaster, the guardian casts the stored spell with any parameters
      set by the original caster, requiring no components. When the spell is
      cast or a new spell is stored, any previously stored spell is lost.
    Multiattack.
      The guardian makes two fist attacks.
    Fist.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 11 (2d6
      + 4) bludgeoning damage.
    """
    name = "Shield Guardian"
    description = "Large construct, unaligned"
    challenge_rating = 7
    armor_class = 17
    skills = ""
    senses = "Blindsight 10 ft., Darkvision 60 ft., Passive Perception 10"
    languages = "understands commands given in any language but can't speak"
    strength = Ability(18)
    dexterity = Ability(8)
    constitution = Ability(18)
    intelligence = Ability(7)
    wisdom = Ability(10)
    charisma = Ability(3)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 142
    hit_dice = "15d10"
    spells = []


class Shrieker(Monster):
    """
    False Appearance.
      While the shrieker remains motionless, it is indistinguishable from an
      ordinary fungus.
    Shriek.
      When bright light or a creature is within 30 feet of the shrieker, it
      emits a shriek audible within 300 feet of it. The shrieker continues
      to shriek until the disturbance moves out of range and for 1d4 of the
      shrieker's turns afterward
    """
    name = "Shrieker"
    description = "Medium plant, unaligned"
    challenge_rating = 0
    armor_class = 5
    skills = ""
    senses = "Blindsight 30 ft. (blind beyond this radius), Passive Perception 6"
    languages = ""
    strength = Ability(1)
    dexterity = Ability(1)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(3)
    charisma = Ability(1)
    speed = 0
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 13
    hit_dice = "3d8"
    spells = []


class SilverDragonWyrmling(Monster):
    """
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 9 (1d10
      + 4) piercing damage.
    Breath Weapons.
      The dragon uses one of the following breath weapons.
      
      Cold Breath. The dragon exhales an icy blast in a 15-foot cone. Each
      creature in that area must make a DC 13 Constitution saving throw,
      taking 18 (4d8) cold damage on a failed save, or half as much damage
      on a successful one.
      
      Paralyzing Breath. The dragon exhales paralyzing gas in a 15-foot
      cone. Each creature in that area must succeed on a DC 13 Constitution
      saving throw or be paralyzed for 1 minute. A creature can repeat the
      saving throw at the end of each of its turns, ending the effect on
      itself on a success.
    """
    name = "Silver Dragon Wyrmling"
    description = "Medium dragon, lawful good"
    challenge_rating = 2
    armor_class = 17
    skills = "Perception +4, Stealth +2"
    senses = "Blindsight 10 ft., Darkvision 60 ft., Passive Perception 14"
    languages = "Draconic"
    strength = Ability(19)
    dexterity = Ability(10)
    constitution = Ability(17)
    intelligence = Ability(12)
    wisdom = Ability(11)
    charisma = Ability(15)
    speed = 30
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 45
    hit_dice = "6d8"
    spells = []


class Skeleton(Monster):
    """
    Shortsword.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 5 (1d6 +
      2) piercing damage.
    Shortbow.
      Ranged Weapon Attack: +4 to hit, range 80/320 ft., one target. Hit: 5
      (1d6 + 2) piercing damage.
    """
    name = "Skeleton"
    description = "Medium undead, lawful evil"
    challenge_rating = 0.25
    armor_class = 13
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 9"
    languages = "understands all languages it spoke in life but can't speak"
    strength = Ability(10)
    dexterity = Ability(14)
    constitution = Ability(15)
    intelligence = Ability(6)
    wisdom = Ability(8)
    charisma = Ability(5)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 13
    hit_dice = "2d8"
    spells = []


class Solar(Monster):
    """
    Angelic Weapons.
      The solar's weapon attacks are magical. When the solar hits with any
      weapon, the weapon deals an extra 6d8 radiant damage (included in the
      attack).
    Divine Awareness.
      The solar knows if it hears a lie.
    Innate Spellcasting.
      The solar's spell casting ability is Charisma (spell save DC 25). It
      can innately cast the following spells, requiring no material
      components:
      
      At will: detect evil and good, invisibility (self only)
      
      3/day each: blade barrier, dispel evil and good, resurrection
      
      1/day each: commune, control weather
    Magic Resistance.
      The solar has advantage on saving throws against spells and other
      magical effects.
    Multiattack.
      The solar makes two greatsword attacks.
    Greatsword.
      Melee Weapon Attack: +15 to hit, reach 5 ft., one target. Hit: 22 (4d6
      + 8) slashing damage plus 27 (6d8) radiant damage.
    Slaying Longbow.
      Ranged Weapon Attack: +13 to hit, range 150/600 ft., one target. Hit:
      15 (2d8 + 6) piercing damage plus 27 (6d8) radiant damage. If the
      target is a creature that has 190 hit points or fewer, it must succeed
      on a DC 15 Constitution saving throw or die.
    Flying Sword.
      The solar releases its greatsword to hover magically in an unoccupied
      space within 5 ft. of it. If the solar can see the sword, the solar
      can mentally command it as a bonus action to fly up to 50 ft. and
      either make one attack against a target or return to the solar's
      hands. If the hovering sword is targeted by any effect, the solar is
      considered to be holding it. The hovering sword falls if the solar
      dies.
    Healing Touch.
      The solar touches another creature. The target magically regains 40
      (8d8 + 4) hit points and is freed from any curse, disease, poison,
      blindness, or deafness.
    """
    name = "Solar"
    description = "Large celestial, lawful good"
    challenge_rating = 21
    armor_class = 21
    skills = "Perception +14"
    senses = "Truesight 120 ft., Passive Perception 24"
    languages = "all, telepathy 120 ft."
    strength = Ability(26)
    dexterity = Ability(22)
    constitution = Ability(26)
    intelligence = Ability(25)
    wisdom = Ability(25)
    charisma = Ability(30)
    speed = 50
    swim_speed = 0
    fly_speed = 150
    climb_speed = 0
    hp_max = 243
    hit_dice = "18d10"
    spells = []


class Specter(Monster):
    """
    Incorporeal Movement.
      The specter can move through other creatures and objects as if they
      were difficult terrain. It takes 5 (1d10) force damage if it ends its
      turn inside an object.
    Sunlight Sensitivity.
      While in sunlight, the specter has disadvantage on attack rolls, as
      well as on Wisdom (Perception) checks that rely on sight.
    Life Drain.
      Melee Spell Attack: +4 to hit, reach 5 ft., one creature. Hit: 10
      (3d6) necrotic damage. The target must succeed on a DC 10 Constitution
      saving throw or its hit point maximum is reduced by an amount equal to
      the damage taken. This reduction lasts until the creature finishes a
      long rest. The target dies if this effect reduces its hit point
      maximum to 0.
    """
    name = "Specter"
    description = "Medium undead, chaotic evil"
    challenge_rating = 1
    armor_class = 12
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "understands all languages it knew in life but can't speak"
    strength = Ability(1)
    dexterity = Ability(14)
    constitution = Ability(11)
    intelligence = Ability(10)
    wisdom = Ability(10)
    charisma = Ability(11)
    speed = 0
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class Spider(Monster):
    """
    Spider Climb.
      The spider can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Web Sense.
      While in contact with a web, the spider knows the exact location of
      any other creature in contact with the same web.
    Web Walker.
      The spider ignores movement restrictions caused by webbing.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one creature. Hit: 1
      piercing damage, and the target must succeed on a DC 9 Constitution
      saving throw or take 2 (1d4) poison damage.
    """
    name = "Spider"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 12
    skills = "Stealth +4"
    senses = "Darkvision 30 ft., Passive Perception 12"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(14)
    constitution = Ability(8)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(2)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 20
    hp_max = 1
    hit_dice = "1d4"
    spells = []


class SpiritNaga(Monster):
    """
    Rejuvenation.
      If it dies, the naga returns to life in 1d6 days and regains all its
      hit points. Only a wish spell can prevent this trait from functioning.
    Spellcasting.
      The naga is a 10th-level spellcaster. Its spellcasting ability is
      Intelligence (spell save DC 14, +6 to hit with spell attacks), and it
      needs only verbal components to cast its spells. It has the following
      wizard spells prepared:
      
  
      
      - Cantrips (at will): mage hand, minor illusion, ray of frost
      
      - 1st level (4 slots): charm person, detect magic, sleep
      
      - 2nd level (3 slots): detect thoughts, hold person
      
      - 3rd level (3 slots): lightning bolt, water breathing
      
      - 4th level (3 slots): blight, dimension door
      
      - 5th level (2 slots): dominate person
    Bite.
      Melee Weapon Attack: +7 to hit, reach 10 ft., one creature. Hit: 7
      (1d6 + 4) piercing damage, and the target must make a DC 13
      Constitution saving throw, taking 31 (7d8) poison damage on a failed
      save, or half as much damage on a successful one.
    """
    name = "Spirit Naga"
    description = "Large monstrosity, chaotic evil"
    challenge_rating = 8
    armor_class = 15
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 12"
    languages = "Abyssal, Common"
    strength = Ability(18)
    dexterity = Ability(17)
    constitution = Ability(14)
    intelligence = Ability(16)
    wisdom = Ability(15)
    charisma = Ability(16)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 75
    hit_dice = "10d10"
    spells = ["mage hand", "minor illusion", "ray of frost", "charm person", "detect magic", "sleep", "detect thoughts", "hold person", "lightning bolt", "water breathing", "blight", "dimension door", "dominate person"]


class Sprite(Monster):
    """
    Longsword.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one target. Hit: 1
      slashing damage.
    Shortbow.
      Ranged Weapon Attack: +6 to hit, range 40/160 ft., one target. Hit: 1
      piercing damage, and the target must succeed on a DC 10 Constitution
      saving throw or become poisoned for 1 minute. If its saving throw
      result is 5 or lower, the poisoned target falls unconscious for the
      same duration, or until it takes damage or another creature takes an
      action to shake it awake.
    Heart Sight.
      The sprite touches a creature and magically knows the creature's
      current emotional state. If the target fails a DC 10 Charisma saving
      throw, the sprite also knows the creature's alignment. Celestials,
      fiends, and undead automatically fail the saving throw.
    Invisibility.
      The sprite magically turns invisible until it attacks or casts a
      spell, or until its concentration ends (as if concentrating on a
      spell). Any equipment the sprite wears or carries is invisible with
      it.
    """
    name = "Sprite"
    description = "Tiny fey, neutral good"
    challenge_rating = 0.25
    armor_class = 15
    skills = "Perception +3, Stealth +8"
    senses = "Passive Perception 13"
    languages = "Common, Elvish, Sylvan"
    strength = Ability(3)
    dexterity = Ability(18)
    constitution = Ability(10)
    intelligence = Ability(14)
    wisdom = Ability(13)
    charisma = Ability(11)
    speed = 10
    swim_speed = 0
    fly_speed = 40
    climb_speed = 0
    hp_max = 2
    hit_dice = "1d4"
    spells = []


class Spy(Monster):
    """
    Cunning Action.
      On each of its turns, the spy can use a bonus action to take the Dash,
      Disengage, or Hide action.
    Sneak Attack (1/Turn).
      The spy deals an extra 7 (2d6) damage when it hits a target with a
      weapon attack and has advantage on the attack roll, or when the target
      is within 5 ft. of an ally of the spy that isn't incapacitated and the
      spy doesn't have disadvantage on the attack roll.
    Multiattack.
      The spy makes two melee attacks.
    Shortsword.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 5 (1d6 +
      2) piercing damage.
    Hand Crossbow.
      Ranged Weapon Attack: +4 to hit, range 30/120 ft., one target. Hit: 5
      (1d6 + 2) piercing damage.
    """
    name = "Spy"
    description = "Medium humanoid, any alignment"
    challenge_rating = 1
    armor_class = 12
    skills = "Deception +5, Insight +4, Investigation +5, Perception +6, Persuasion +5, Stealth +4"
    senses = "Passive Perception 16"
    languages = "any two languages"
    strength = Ability(10)
    dexterity = Ability(15)
    constitution = Ability(10)
    intelligence = Ability(12)
    wisdom = Ability(14)
    charisma = Ability(16)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 27
    hit_dice = "6d8"
    spells = []


class SteamMephit(Monster):
    """
    Death Burst.
      When the mephit dies, it explodes in a cloud of steam. Each creature
      within 5 ft. of the mephit must succeed on a DC 10 Dexterity saving
      throw or take 4 (1d8) fire damage.
    Innate Spellcasting.
      The mephit can innately cast blur, requiring no material components.
      Its innate spellcasting ability is Charisma.
    Claws.
      Melee Weapon Attack: +2 to hit, reach 5 ft., one creature. Hit: 2
      (1d4) slashing damage plus 2 (1d4) fire damage.
    Steam Breath (Recharge 6).
      The mephit exhales a 15-foot cone of scalding steam. Each creature in
      that area must succeed on a DC 10 Dexterity saving throw, taking 4
      (1d8) fire damage on a failed save, or half as much damage on a
      successful one.
    Variant: Summon Mephits.
      The mephit has a 25 percent chance of summoning 1d4 mephits of its
      kind. A summoned mephit appears in an unoccupied space within 60 feet
      of its summoner, acts as an ally of its summoner, and can't summon
      other mephits. It remains for 1 minute, until it or its summoner dies,
      or until its summoner dismisses it as an action.
    """
    name = "Steam Mephit"
    description = "Small elemental, neutral evil"
    challenge_rating = 0.25
    armor_class = 10
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 10"
    languages = "Aquan, Ignan"
    strength = Ability(5)
    dexterity = Ability(11)
    constitution = Ability(10)
    intelligence = Ability(11)
    wisdom = Ability(10)
    charisma = Ability(12)
    speed = 30
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 21
    hit_dice = "6d6"
    spells = []


class Stirge(Monster):
    """
    Blood Drain.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one creature. Hit: 5 (1d4
      + 3) piercing damage, and the stirge attaches to the target. While
      attached, the stirge doesn't attack. Instead, at the start of each of
      the stirge's turns, the target loses 5 (1d4 + 3) hit points due to
      blood loss.
      
      The stirge can detach itself by spending 5 feet of its movement. It
      does so after it drains 10 hit points of blood from the target or the
      target dies. A creature, including the target, can use its action to
      detach the stirge.
    """
    name = "Stirge"
    description = "Tiny beast, unaligned"
    challenge_rating = 0.125
    armor_class = 14
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 9"
    languages = ""
    strength = Ability(4)
    dexterity = Ability(16)
    constitution = Ability(11)
    intelligence = Ability(2)
    wisdom = Ability(8)
    charisma = Ability(6)
    speed = 10
    swim_speed = 0
    fly_speed = 40
    climb_speed = 0
    hp_max = 2
    hit_dice = "1d4"
    spells = []


class StoneGiant(Monster):
    """
    Stone Camouflage.
      The giant has advantage on Dexterity (Stealth) checks made to hide in
      rocky terrain.
    Multiattack.
      The giant makes two greatclub attacks.
    Greatclub.
      Melee Weapon Attack: +9 to hit, reach 15 ft., one target. Hit: 19 (3d8
      + 6) bludgeoning damage.
    Rock.
      Ranged Weapon Attack: +9 to hit, range 60/240 ft., one target. Hit: 28
      (4d10 + 6) bludgeoning damage. If the target is a creature, it must
      succeed on a DC 17 Strength saving throw or be knocked prone.
    """
    name = "Stone Giant"
    description = "Huge giant, neutral"
    challenge_rating = 7
    armor_class = 17
    skills = "Athletics +12, Perception +4"
    senses = "Darkvision 60 ft., Passive Perception 14"
    languages = "Giant"
    strength = Ability(23)
    dexterity = Ability(15)
    constitution = Ability(20)
    intelligence = Ability(10)
    wisdom = Ability(12)
    charisma = Ability(9)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 126
    hit_dice = "11d12"
    spells = []


class StoneGolem(Monster):
    """
    Immutable Form.
      The golem is immune to any spell or effect that would alter its form.
    Magic Resistance.
      The golem has advantage on saving throws against spells and other
      magical effects.
    Magic Weapons.
      The golem's weapon attacks are magical.
    Multiattack.
      The golem makes two slam attacks.
    Slam.
      Melee Weapon Attack: +10 to hit, reach 5 ft., one target. Hit: 19 (3d8
      + 6) bludgeoning damage.
    Slow.
      The golem targets one or more creatures it can see within 10 ft. of
      it. Each target must make a DC 17 Wisdom saving throw against this
      magic. On a failed save, a target can't use reactions, its speed is
      halved, and it can't make more than one attack on its turn. In
      addition, the target can take either an action or a bonus action on
      its turn, not both. These effects last for 1 minute. A target can
      repeat the saving throw at the end of each of its turns, ending the
      effect on itself on a success.
    """
    name = "Stone Golem"
    description = "Large construct, unaligned"
    challenge_rating = 10
    armor_class = 17
    skills = ""
    senses = "Darkvision 120 ft., Passive Perception 10"
    languages = "understands the languages of its creator but can't speak"
    strength = Ability(22)
    dexterity = Ability(9)
    constitution = Ability(20)
    intelligence = Ability(3)
    wisdom = Ability(11)
    charisma = Ability(1)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 178
    hit_dice = "17d10"
    spells = []


class Stonemelder(Monster):
    """Death Burst.
      When the stonemelder dies, it turns to stone and explodes in a
      burst of rock shards, becoming a smoking pile of rubble. Each
      creature within 10 feet of the exploding stonemelder must make a
      DC 14 Dexterity saving throw, taking 11 (2d10) bludgeoning
      damage on a failed save, or half as much damage on a successful
      one.
    Spellcasting.
      The stonemelder is a 7th-level spellcaster. Its spellcasting
      ability is Charisma (spell save DC 13, +5 to hit with spell
      attacks. It knows the following sorcerer spells (an asterisked
      spell is from *Princes of the Apocalypse* appendix B):
      
      - Cantrips (at will): acid splash, blade ward, light mending, mold earth*
      - 1st level (4 slots): expeditious retreat, false life, shield
      - 2nd level (3 slots): Maximilian's earthen grasp,* shatter
      - 3rd level (3 slots): erupting earth,* meld into stone
      - 4th level (1 slots): stoneskin
      
    Black Earth Rod.
      *Melee Weapon Attack:* +5 to hit, reach 5 ft., one
      target. *Hit:* 5 (1d6 + 2) bludgeoning damage. The stonemelder
      can also expend a spell slot to deal extra damage, dealing 2d8
      bludgeoning damage for a 1st level slot, plus an additional 1d8
      for each level of the slot above 1st.
    
    Stonemelders are spellcasters who gain their power from Ogrémoch,
    using elemental earth magic to sheathe their bodies in carapaces
    of rock. They are elite champions of the cult and answer only to
    the cult leaders. Not even the Black Earth priests tell
    stonemelders what to do.
    
    Each stonemelder carries a weapon known as a Black Earth rod. Such
    a rod is like an ordinary mace, but in the hands of a stonemelder,
    it serves as a conduit for Ogrémoch's wrath.

    """
    name = "Stonemelder"
    description = "Medium humanoid (human), neutral evil"
    challenge_rating = 4
    armor_class = 17
    skills = "Intimidation +5, Perception +2"
    senses = "termorsense 30 ft., passive Perception 12"
    languages = "Common, Terran"
    strength = Ability(15)
    dexterity = Ability(10)
    constitution = Ability(16)
    intelligence = Ability(12)
    wisdom = Ability(11)
    charisma = Ability(17)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 75
    hit_dice = "10d8 + 30"
    spells = ["acid splash", "blade ward", "light mending", "mold earth*",
              "expeditious retreat", "false life", "shield",
              "Maximilian's earthen grasp*", "shatter" "erupting earth*",
              "meld into stone", "stoneskin"]


class StormGiant(Monster):
    """
    Amphibious.
      The giant can breathe air and water.
    Innate Spellcasting.
      The giant's innate spellcasting ability is Charisma (spell save DC
      17). It can innately cast the following spells, requiring no material
      components:
      
  
      
      At will: detect magic, feather fall, levitate, light
      
      3/day each: control weather, water breathing
    Multiattack.
      The giant makes two greatsword attacks.
    Greatsword.
      Melee Weapon Attack: +14 to hit, reach 10 ft., one target. Hit: 30
      (6d6 + 9) slashing damage.
    Rock.
      Ranged Weapon Attack: +14 to hit, range 60/240 ft., one target. Hit:
      35 (4d12 + 9) bludgeoning damage.
    Lightning Strike.
      The giant hurls a magical lightning bolt at a point it can see within
      500 feet of it. Each creature within 10 feet of that point must make a
      DC 17 Dexterity saving throw, taking 54 (12d8) lightning damage on a
      failed save, or half as much damage on a successful one.
    """
    name = "Storm Giant"
    description = "Huge giant, chaotic good"
    challenge_rating = 13
    armor_class = 16
    skills = "Arcana +8, Athletics +14, History +8, Perception +9"
    senses = "Passive Perception 19"
    languages = "Common, Giant"
    strength = Ability(29)
    dexterity = Ability(14)
    constitution = Ability(20)
    intelligence = Ability(16)
    wisdom = Ability(18)
    charisma = Ability(18)
    speed = 50
    swim_speed = 50
    fly_speed = 0
    climb_speed = 0
    hp_max = 230
    hit_dice = "20d12"
    spells = []


class SuccubusIncubus(Monster):
    """
    Telepathic Bond.
      The fiend ignores the range restriction on its telepathy when
      communicating with a creature it has charmed. The two don't even need
      to be on the same plane of existence.
    Shapechanger.
      The fiend can use its action to polymorph into a Small or Medium
      humanoid, or back into its true form. Without wings, the fiend loses
      its flying speed. Other than its size and speed, its statistics are
      the same in each form. Any equipment it is wearing or carrying isn't
      transformed. It reverts to its true form if it dies.
    Claw (Fiend Form Only).
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 6 (1d6 +
      3) slashing damage.
    Charm.
      One humanoid the fiend can see within 30 feet of it must succeed on a
      DC 15 Wisdom saving throw or be magically charmed for 1 day. The
      charmed target obeys the fiend's verbal or telepathic commands. If the
      target suffers any harm or receives a suicidal command, it can repeat
      the saving throw, ending the effect on a success. If the target
      successfully saves against the effect, or if the effect on it ends,
      the target is immune to this fiend's Charm for the next 24 hours.
      
      The fiend can have only one target charmed at a time. If it charms
      another, the effect on the previous target ends.
    Draining Kiss.
      The fiend kisses a creature charmed by it or a willing creature. The
      target must make a DC 15 Constitution saving throw against this magic,
      taking 32 (5d10 + 5) psychic damage on a failed save, or half as much
      damage on a successful one. The target's hit point maximum is reduced
      by an amount equal to the damage taken. This reduction lasts until the
      target finishes a long rest. The target dies if this effect reduces
      its hit point maximum to 0.
    Etherealness.
      The fiend magically enters the Ethereal Plane from the Material Plane,
      or vice versa.
    """
    name = "Succubus/Incubus"
    description = "Medium fiend, neutral evil"
    challenge_rating = 4
    armor_class = 15
    skills = "Deception +9, Insight +5, Perception +5, Persuasion +9, Stealth +7"
    senses = "Darkvision 60 ft., Passive Perception 15"
    languages = "Abyssal, Common, Infernal, telepathy 60 ft."
    strength = Ability(8)
    dexterity = Ability(17)
    constitution = Ability(13)
    intelligence = Ability(15)
    wisdom = Ability(12)
    charisma = Ability(20)
    speed = 30
    swim_speed = 0
    fly_speed = 60
    climb_speed = 0
    hp_max = 66
    hit_dice = "12d8"
    spells = []


class SwarmOfBats(Monster):
    """
    Echolocation.
      The swarm can't use its blindsight while deafened.
    Keen Hearing.
      The swarm has advantage on Wisdom (Perception) checks that rely on
      hearing.
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny bat. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +4 to hit, reach 0 ft., one creature in the
      swarm's space. Hit: 5 (2d4) piercing damage, or 2 (1d4) piercing
      damage if the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Bats"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.25
    armor_class = 12
    skills = ""
    senses = "Blindsight 60 ft., Passive Perception 11"
    languages = ""
    strength = Ability(5)
    dexterity = Ability(15)
    constitution = Ability(10)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(4)
    speed = 0
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class SwarmOfBats(Monster):
    """Echolocation.
      The swarm can't use its blindsight while deafened.
    Keen Hearing.
      The swarm has advantage on Wisdom (Perception) checks that rely on hearing.
    Swarm.
      The swarm can occupy another creature's space and vice versa,
      and the swarm can move through any opening large enough for a
      Tiny bat. The swarm can't regain hit points or gain temporary
      hit points.
    Bites.
      Melee Weapon Attack: +4 to hit, reach 0 ft., one creature in the
      swarm's space. Hit: 5 (2d4) piercing damage, or 2 (1d4) piercing
      damage if the swarm has half of its hit points or fewer.
    
    """
    name = "Swarm of Bats"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.25
    armor_class = 12
    skills = ""
    senses = "Blindsight 60 ft., Passive Perception 11"
    languages = ""
    strength = Ability(5)
    dexterity = Ability(15)
    constitution = Ability(10)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(4)
    speed = 0
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 22
    hit_dice = "5d8"


class SwarmOfBeetles(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny insect. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +3 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 10 (4d4) piercing damage, or 5 (2d4) piercing damage if
      the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Beetles"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 8"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(1)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 20
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class SwarmOfCentipedes(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny insect. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +3 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 10 (4d4) piercing damage, or 5 (2d4) piercing damage if
      the swarm has half of its hit points or fewer.
      
      A creature reduced to 0 hit points by a swarm of centipedes is stable
      but poisoned for 1 hour, even after regaining hit points, and
      paralyzed while poisoned in this way.
    """
    name = "Swarm of Centipedes"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 8"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(1)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 20
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class SwarmOfInsects(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny insect. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +3 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 10 (4d4) piercing damage, or 5 (2d4) piercing damage if
      the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Insects"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 8"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(1)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 20
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class SwarmOfPoisonousSnakes(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny snake. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +6 to hit, reach 0 ft., one creature in the
      swarm's space. Hit: 7 (2d6) piercing damage, or 3 (1d6) piercing
      damage if the swarm has half of its hit points or fewer. The target
      must make a DC 10 Constitution saving throw, taking 14 (4d6) poison
      damage on a failed save, or half as much damage on a successful one.
    """
    name = "Swarm of Poisonous Snakes"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 2
    armor_class = 14
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 10"
    languages = ""
    strength = Ability(8)
    dexterity = Ability(18)
    constitution = Ability(11)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(3)
    speed = 30
    swim_speed = 30
    fly_speed = 0
    climb_speed = 0
    hp_max = 36
    hit_dice = "8d8"
    spells = []


class SwarmOfQuippers(Monster):
    """
    Blood Frenzy.
      The swarm has advantage on melee attack rolls against any creature
      that doesn't have all its hit points.
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny quipper.
      The swarm can't regain hit points or gain temporary hit points.
    Water Breathing.
      The swarm can breathe only underwater.
    Bites.
      Melee Weapon Attack: +5 to hit, reach 0 ft., one creature in the
      swarm's space. Hit: 14 (4d6) piercing damage, or 7 (2d6) piercing
      damage if the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Quippers"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 1
    armor_class = 13
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 8"
    languages = ""
    strength = Ability(13)
    dexterity = Ability(16)
    constitution = Ability(9)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(2)
    speed = 0
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 28
    hit_dice = "8d8"
    spells = []


class SwarmOfRats(Monster):
    """
    Keen Smell.
      The swarm has advantage on Wisdom (Perception) checks that rely on
      smell.
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny rat. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +2 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 7 (2d6) piercing damage, or 3 (1d6) piercing damage if the
      swarm has half of its hit points or fewer.
    """
    name = "Swarm of Rats"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.25
    armor_class = 10
    skills = ""
    senses = "Darkvision 30 ft., Passive Perception 10"
    languages = ""
    strength = Ability(9)
    dexterity = Ability(11)
    constitution = Ability(9)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(3)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 24
    hit_dice = "7d8"
    spells = []


class SwarmOfRavens(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny raven. The
      swarm can't regain hit points or gain temporary hit points.
    Beaks.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target in the swarm's
      space. Hit: 7 (2d6) piercing damage, or 3 (1d6) piercing damage if the
      swarm has half of its hit points or fewer.
    """
    name = "Swarm of Ravens"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.25
    armor_class = 12
    skills = ""
    senses = "Passive Perception 15"
    languages = ""
    strength = Ability(6)
    dexterity = Ability(14)
    constitution = Ability(8)
    intelligence = Ability(3)
    wisdom = Ability(12)
    charisma = Ability(6)
    speed = 10
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 24
    hit_dice = "7d8"
    spells = []


class SwarmOfSpiders(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny insect. The
      swarm can't regain hit points or gain temporary hit points.
    Spider Climb.
      The swarm can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Web Sense.
      While in contact with a web, the swarm knows the exact location of any
      other creature in contact with the same web.
    Web Walker.
      The swarm ignores movement restrictions caused by webbing.
    Bites.
      Melee Weapon Attack: +3 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 10 (4d4) piercing damage, or 5 (2d4) piercing damage if
      the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Spiders"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 8"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(1)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 20
    hp_max = 22
    hit_dice = "5d8"
    spells = []


class SwarmOfWasps(Monster):
    """
    Swarm.
      The swarm can occupy another creature's space and vice versa, and the
      swarm can move through any opening large enough for a Tiny insect. The
      swarm can't regain hit points or gain temporary hit points.
    Bites.
      Melee Weapon Attack: +3 to hit, reach 0 ft., one target in the swarm's
      space. Hit: 10 (4d4) piercing damage, or 5 (2d4) piercing damage if
      the swarm has half of its hit points or fewer.
    """
    name = "Swarm of Wasps"
    description = "Medium swarm of Tiny beasts, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = ""
    senses = "Blindsight 10 ft., Passive Perception 8"
    languages = ""
    strength = Ability(3)
    dexterity = Ability(13)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(7)
    charisma = Ability(1)
    speed = 5
    swim_speed = 0
    fly_speed = 30
    climb_speed = 0
    hp_max = 22
    hit_dice = "5d8"
    spells = []



from dungeonsheets.spells.spells import Spell


class VampiricTouch(Spell):
    """The touch of your shadow-wreathed hand can siphon force from others to heal your
    wounds. Make a melee spell attack against a creature within your reach. On a
    hit, the target takes 3d6 necrotic damage, and you regain hit points equal to
    half the amount of necrotic damage dealt. Until the spell ends, you can make the
    attack again on each of your turns as an action.

    **At Higher Levels:** When you
    cast this spell using a spell slot of 4th level or higher, the damage increases
    by 1d6 for each slot level above 3rd.
    """

    name = "Vampiric Touch"
    level = 3
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Warlock", "Wizard")


class ViciousMockery(Spell):
    """You unleash a string of insults laced with subtle enchantments at a creature you
    can see within range.
    If the target can hear you (thought it need not
    understand you), it must succeed on a Wisdom saving throw or take 1d4 psychic
    damage and have disadvantage on the next attack roll it makes before the end of
    its next turn.

    **At Higher Levels:** This spell's damage increases by 1d4 when you
    reach 5th level (2d4), 11th level (3d4), and 17th level (4d4).
    """

    name = "Vicious Mockery"
    level = 0
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard",)


class VitriolicSphere(Spell):
    """You point at a location within range, and a glowing, 1-foot-diameter ball of
    emerald acid streaks there and explodes in a 20-foot-radius sphere. Each
    creature in that area must make a Dexterity saving throw. On a failed save, a
    creature takes 10d4 acid damage and another 5d4 acid damage at the end of its
    next turn. On a successful save, a creature takes half the initial damage and no
    damage at the end of its next turn.

    **At Higher Levels:** When you cast this spell
    using a spell slot of 5th level or higher, the initial damage increases by 2d4
    for each slot level above 4th.
    """

    name = "Vitriolic Sphere"
    level = 4
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S", "M")
    materials = "A drop of giant slug bile"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")



"""Tools for describing a player character."""
import os
import re
import warnings
import math
from types import ModuleType
from typing import Sequence, Union, MutableMapping

import jinja2

from dungeonsheets import (
    armor,
    background,
    classes,
    exceptions,
    features,
    infusions,
    magic_items,
    monsters,
    race,
    spells,
    weapons,
)
from dungeonsheets.content_registry import find_content
from dungeonsheets.weapons import Weapon
from dungeonsheets.content import Creature
from dungeonsheets.dice import combine_dice


dice_re = re.compile(r"(\d+)d(\d+)")

__all__ = (
    "Artificer",
    "Barbarian",
    "Bard",
    "Cleric",
    "Character",
    "Druid",
    "Fighter",
    "Monk",
    "Paladin",
    "Ranger",
    "Rogue",
    "Sorceror",
    "Warlock",
    "Wizard",
)

multiclass_spellslots_by_level = {
    # char_lvl: (cantrips, 1st, 2nd, 3rd, ...)
    1: (0, 2, 0, 0, 0, 0, 0, 0, 0, 0),
    2: (0, 3, 0, 0, 0, 0, 0, 0, 0, 0),
    3: (0, 4, 2, 0, 0, 0, 0, 0, 0, 0),
    4: (0, 4, 3, 0, 0, 0, 0, 0, 0, 0),
    5: (0, 4, 3, 2, 0, 0, 0, 0, 0, 0),
    6: (0, 4, 3, 3, 0, 0, 0, 0, 0, 0),
    7: (0, 4, 3, 3, 1, 0, 0, 0, 0, 0),
    8: (0, 4, 3, 3, 2, 0, 0, 0, 0, 0),
    9: (0, 4, 3, 3, 3, 1, 0, 0, 0, 0),
    10: (0, 4, 3, 3, 3, 2, 0, 0, 0, 0),
    11: (0, 4, 3, 3, 3, 2, 1, 0, 0, 0),
    12: (0, 4, 3, 3, 3, 2, 1, 0, 0, 0),
    13: (0, 4, 3, 3, 3, 2, 1, 1, 0, 0),
    14: (0, 4, 3, 3, 3, 2, 1, 1, 0, 0),
    15: (0, 4, 3, 3, 3, 2, 1, 1, 1, 0),
    16: (0, 4, 3, 3, 3, 2, 1, 1, 1, 0),
    17: (0, 4, 3, 3, 3, 2, 1, 1, 1, 1),
    18: (0, 4, 3, 3, 3, 3, 1, 1, 1, 1),
    19: (0, 4, 3, 3, 3, 3, 2, 1, 1, 1),
    20: (0, 4, 3, 3, 3, 3, 2, 2, 1, 1),
}


class Character(Creature):
    """A generic player character."""
    # Character-specific
    name = "Unknown Hero"
    player_name = ""
    xp = 0
    # Extra hit points info, for characters only
    hp_current = None
    hp_temp = 0
    # Base stats (ability scores)
    inspiration = False
    attacks_and_spellcasting = ""
    class_list = list()
    _background = None

    # Characteristics
    personality_traits = (
        "TODO: Describe how your character behaves, interacts with others"
    )
    ideals = "TODO: Describe what values your character believes in."
    bonds = "TODO: Describe your character's commitments or ongoing quests."
    flaws = "TODO: Describe your character's interesting flaws."
    features_and_traits = "Describe any other features and abilities."

    _proficiencies_text = list()

    # Appearance
    portrait = False
    age = 0
    height = ""
    weight = ""
    eyes = ""
    skin = ""
    hair = ""
    # Background
    allies = ""
    faction_name = ""
    # faction_symbol = placeholder not sure how to implement
    backstory = ""
    other_feats_traits = ""
    treasure = ""

    def __init__(
        self,
        classes: Sequence = [],
        levels: Sequence[int] = [],
        subclasses: Sequence = [],
        **attrs,
    ):
        """Create a new character from attributes *attrs*.

        **Multiclassing** can be accomplished by a list of class names
        *classes*, and a list of class levels *levels*.

        Parameters
        ==========
        classes
          Strings with class names, or character class definitions
          representing the characters various D&D classes.
        levels
          The class levels for each corresponding class entry in
          *classes*.
        subclasses
          Subclasses that apply for this character.
        **attrs
          Additional keyword parameters to set as attributes for this
          character.

        """
        super(Character, self).__init__()
        self.clear()
        # make sure class, race, background are set first
        my_classes = classes
        my_levels = levels
        my_subclasses = subclasses
        # backwards compatability
        if len(my_classes) == 0:
            if "class" in attrs:
                my_classes = [attrs.pop("class")]
                my_levels = [attrs.pop("level", 1)]
                my_subclasses = [attrs.pop("subclass", None)]
            else:  # if no classes or levels given, default to Lvl 1 Fighter
                my_classes = ["Fighter"]
                my_levels = [1]
                my_subclasses = [None]
        # Generate the list of class objects
        self.add_classes(
            my_classes,
            my_levels,
            my_subclasses,
            feature_choices=attrs.get("feature_choices", []),
        )
        # parse race and background
        self.race = attrs.pop("race", None)
        self.background = attrs.pop("background", None)
        # parse all other attributes
        self.set_attrs(**attrs)
        self.__set_max_hp(attrs.get("hp_max", None))

    def clear(self):
        # reset class-defined items
        self.class_list = list()
        self._weapons = list()
        self.magic_items = list()
        self._saving_throw_proficiencies = tuple()
        self.other_weapon_proficiencies = tuple()
        self.skill_proficiencies = list()
        self.skill_expertise = list()
        self._proficiencies_text = list()
        self._spells = list()
        self._spells_prepared = list()
        self.infusions = list()
        self.custom_features = list()
        self.feature_choices = list()

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"<{self.class_name}: {self.name}>"

    def add_class(
        self,
        cls: (classes.CharClass, type, str),
        level: (int, str),
        subclass=None,
        feature_choices: Sequence = [],
    ):
        """Add a class, level, and subclass the character has attained."""
        if isinstance(cls, str):
            cls = cls.strip().title().replace(" ", "")
            try:
                cls = getattr(classes, cls)
            except AttributeError:
                raise AttributeError(
                    "class was not recognized from classes.py: {:s}".format(cls)
                )
        if isinstance(level, str):
            level = int(level)
        self.class_list.append(
            cls(level, owner=self, subclass=subclass, feature_choices=feature_choices)
        )

    def add_classes(
        self,
        classes_list: Sequence[Union[str, classes.CharClass]] = [],
        levels: Sequence[Union[int, float, str]] = [],
        subclasses: Sequence = [],
        feature_choices: Sequence = [],
    ):
        """Add several classes, levels, etc.

        The lists can also be single values for a single class
        character.

        """
        if isinstance(classes_list, str):
            classes_list = [classes_list]
        if (
            isinstance(levels, int)
            or isinstance(levels, float)
            or isinstance(levels, str)
        ):
            levels = [levels]
        if len(levels) == 0:
            levels = [1] * len(classes_list)
        if isinstance(subclasses, str):
            subclasses = [subclasses]
        if len(subclasses) == 0:
            subclasses = [None] * len(classes_list)
        assert len(classes_list) == len(
            levels
        ), "the length of classes {:d} does not match length of levels {:d}".format(
            len(classes), len(levels)
        )
        assert len(classes_list) == len(
            subclasses
        ), "the length of classes {:d} does not match length of subclasses {:d}".format(
            len(classes_list), len(subclasses)
        )
        for cls, lvl, sub in zip(classes_list, levels, subclasses):
            params = {}
            params["feature_choices"] = feature_choices
            self.add_class(cls=cls, level=lvl, subclass=sub, **params)

    @property
    def race(self):
        return self._race

    @race.setter
    def race(self, newrace):
        if isinstance(newrace, race.Race):
            self._race = newrace
            self._race.owner = self
        elif isinstance(newrace, type) and issubclass(newrace, race.Race):
            self._race = newrace(owner=self)
        elif isinstance(newrace, str):
            try:
                self._race = find_content(newrace, valid_classes=[race.Race])(
                    owner=self
                )
            except exceptions.ContentNotFound:
                msg = f'Race "{newrace}" not defined. Please add it to ``race.py``'
                self._race = race.Race(owner=self)
                warnings.warn(msg)
        elif newrace is None:
            self._race = race.Race(owner=self)

    @property
    def background(self):
        return self._background

    @background.setter
    def background(self, bg):
        if isinstance(bg, background.Background):
            self._background = bg
            self._background.owner = self
        elif isinstance(bg, type) and issubclass(bg, background.Background):
            self._background = bg(owner=self)
        elif isinstance(bg, str):
            try:
                self._background = find_content(
                    bg, valid_classes=[background.Background]
                )(owner=self)
            except exceptions.ContentNotFound:
                msg = (
                    f'Background "{bg}" not defined. Please add it to ``background.py``'
                )
                self._background = background.Background(owner=self)
                warnings.warn(msg)

    @property
    def class_name(self):
        if self.num_classes >= 1:
            return self.primary_class.name
        else:
            return ""

    @property
    def classes_and_levels(self):
        return " / ".join([f"{c.name} {c.level}" for c in self.class_list])

    @property
    def class_names(self):
        return [c.name for c in self.class_list]

    @property
    def levels(self):
        return [c.level for c in self.class_list]

    @property
    def subclasses(self):
        return [c.subclass for c in self.class_list if c.subclass is not None]

    @property
    def level(self):
        return sum(c.level for c in self.class_list)

    @level.setter
    def level(self, new_level):
        self.primary_class.level = new_level
        if self.num_classes > 1:
            warnings.warn(
                "Unable to tell which level to set. Updating "
                "level of primary class {:s}".format(self.primary_class.name)
            )

    @property
    def num_classes(self):
        return len(self.class_list)

    @property
    def has_class(self):
        return self.num_classes > 0

    @property
    def primary_class(self):
        # for now, assume first class given must be primary class
        if self.has_class:
            return self.class_list[0]
        else:
            return None

    def __set_max_hp(self, hp_max):
        """
        Set maximum HP based on value in charlist py or calc from classes
        """
        if hp_max:
            assert isinstance(hp_max, int), hp_max.__class__
            self.hp_max = hp_max
        else:
            const_mod = self.constitution.modifier
            level_one_hp = self.primary_class.hit_dice_faces + const_mod
            self.hp_max = level_one_hp
            for char_cls in self.class_list:
                hp_per_lvl = char_cls.hit_dice_faces / 2 + 1 + const_mod
                levels = char_cls.level
                if char_cls == self.primary_class:
                    levels -= 1
                assert levels >= 0
                self.hp_max += int(hp_per_lvl * levels)

    @property
    def weapon_proficiencies(self):
        wp = set(self.other_weapon_proficiencies)
        if self.num_classes > 0:
            wp |= set(self.primary_class.weapon_proficiencies)
        if self.num_classes > 1:
            for c in self.class_list[1:]:
                wp |= set(c.multiclass_weapon_proficiencies)
        if self.race is not None:
            wp |= set(getattr(self.race, "weapon_proficiencies", ()))
        if self.background is not None:
            wp |= set(getattr(self.background, "weapon_proficiencies", ()))
        return tuple(wp)

    @weapon_proficiencies.setter
    def weapon_proficiencies(self, new_weapons):
        self.other_weapon_proficiencies = tuple(new_weapons)

    @property
    def other_weapon_proficiencies_text(self):
        return tuple(w.name for w in self.other_weapon_proficiencies)

    @property
    def features(self):
        fts = set(self.custom_features)
        fighting_style_defined = False
        set_of_fighting_styles = {
            "Fighting Style (Archery)",
            "Fighting Style (Defense)",
            "Fighting Style (Dueling)",
            "Fighting Style (Great Weapon Fighting)",
            "Fighting Style (Protection)",
            "Fighting Style (Two-Weapon Fighting)",
        }
        for temp_feature in fts:
            fighting_style_defined = temp_feature.name in set_of_fighting_styles
            if fighting_style_defined:
                break

        if not self.has_class:
            return fts
        for c in self.class_list:
            fts |= set(c.features)
            for feature in fts:
                if (
                    fighting_style_defined
                    and feature.name == "Fighting Style (Select One)"
                ):
                    temp_feature = feature
                    fts.remove(temp_feature)
                    break
        if self.race is not None:
            fts |= set(getattr(self.race, "features", ()))
            # some races have level-based features (Ex: Aasimar)
            if hasattr(self.race, "features_by_level"):
                for lvl in range(1, self.level + 1):
                    fts |= set(self.race.features_by_level[lvl])
        if self.background is not None:
            fts |= set(getattr(self.background, "features", ()))

        return sorted(tuple(fts), key=(lambda x: x.name))

    @property
    def custom_features_text(self):
        return tuple([f.name for f in self.custom_features])

    def has_feature(self, feat):
        return any([isinstance(f, feat) for f in self.features])

    @property
    def saving_throw_proficiencies(self):
        if self.primary_class is None:
            return self._saving_throw_proficiencies
        else:
            return (
                self._saving_throw_proficiencies
                or self.primary_class.saving_throw_proficiencies
            )

    @saving_throw_proficiencies.setter
    def saving_throw_proficiencies(self, vals):
        self._saving_throw_proficiencies = vals

    @property
    def spellcasting_classes(self):
        return [c for c in self.class_list if c.is_spellcaster]

    @property
    def spellcasting_classes_excluding_warlock(self):
        return [c for c in self.spellcasting_classes if not type(c) == classes.Warlock]

    @property
    def is_spellcaster(self):
        return len(self.spellcasting_classes) > 0

    def spell_slots(self, spell_level):
        warlock_slots = 0
        for c in self.spellcasting_classes:
            if type(c) is classes.Warlock:
                warlock_slots = c.spell_slots(spell_level)
        if len(self.spellcasting_classes_excluding_warlock) == 0:
            return warlock_slots
        if len(self.spellcasting_classes_excluding_warlock) == 1:
            return (
                self.spellcasting_classes_excluding_warlock[0].spell_slots(spell_level)
                + warlock_slots
            )
        else:
            if spell_level == 0:
                return sum([c.spell_slots(0) for c in self.spellcasting_classes])
            else:
                # compute effective level from PHB pg 164
                eff_level = 0
                for c in self.spellcasting_classes_excluding_warlock:
                    if type(c) in [
                        classes.Bard,
                        classes.Cleric,
                        classes.Druid,
                        classes.Sorceror,
                        classes.Wizard,
                    ]:
                        eff_level += c.level
                    elif type(c) in [classes.Paladin, classes.Ranger]:
                        eff_level += c.level // 2
                    elif type(c) in [classes.Fighter, classes.Rogue]:
                        eff_level += c.level // 3
                    elif type(c) is classes.Artificer:
                        eff_level += math.ceil(c.level / 2)
                if eff_level == 0:
                    return warlock_slots
                else:
                    return (
                        multiclass_spellslots_by_level[eff_level][spell_level]
                        + warlock_slots
                    )

    @property
    def spells(self):
        spells = set(self._spells) | set(self._spells_prepared)
        for f in self.features:
            spells |= set(f.spells_known) | set(f.spells_prepared)
        for c in self.spellcasting_classes:
            spells |= set(c.spells_known) | set(c.spells_prepared)
        if self.race is not None:
            spells |= set(self.race.spells_known) | set(self.race.spells_prepared)
        return sorted(tuple(spells), key=(lambda x: x.name))

    @property
    def spells_prepared(self):
        spells = set(self._spells_prepared)
        for f in self.features:
            spells |= set(f.spells_prepared)
        for c in self.spellcasting_classes:
            spells |= set(c.spells_prepared)
        if self.race is not None:
            spells |= set(self.race.spells_prepared)
        return sorted(tuple(spells), key=(lambda x: x.name))

    def set_attrs(self, **attrs):
        """
        Bulk setting of attributes
        Useful for loading a character from a dictionary
        """
        for attr, val in attrs.items():
            if attr == "dungeonsheets_version":
                pass  # Maybe we'll verify this later?
            elif attr == "weapons":
                if isinstance(val, str):
                    val = [val]
                # Treat weapons specially
                for weap in val:
                    self.wield_weapon(weap)
            elif attr == "magic_items":
                if isinstance(val, str):
                    val = [val]
                for mitem in val:
                    msg = (
                        f'Magic Item "{mitem}" not defined. '
                        "Please add it to ``magic_items.py``"
                    )
                    ThisMagicItem = self._resolve_mechanic(
                        mechanic=mitem,
                        SuperClass=magic_items.MagicItem,
                        warning_message=msg,
                    )
                    self.magic_items.append(ThisMagicItem(owner=self))
            elif attr == "weapon_proficiencies":
                self.other_weapon_proficiencies = ()
                msg = 'Magic Item "{}" not defined. Please add it to ``weapons.py``'
                wps = set(
                    [
                        self._resolve_mechanic(
                            w, SuperClass=weapons.Weapon, warning_message=msg
                        )
                        for w in val
                    ]
                )
                wps -= set(self.weapon_proficiencies)
                self.other_weapon_proficiencies = list(wps)
            elif attr == "armor":
                self.wear_armor(val)
            elif attr == "shield":
                self.wield_shield(val)
            elif attr == "circle":
                if hasattr(self, "Druid"):
                    self.Druid.circle = val
            elif attr == "features":
                if isinstance(val, str):
                    val = [val]
                _features = []
                for f in val:
                    msg = 'Feature "{}" not defined. Please add it to ``features.py``'
                    ThisFeature = self._resolve_mechanic(
                        mechanic=f,
                        SuperClass=features.Feature,
                        warning_message=msg,
                    )
                    _features.append(ThisFeature)
                self.custom_features += tuple(F(owner=self) for F in _features)
            elif (attr == "spells") or (attr == "spells_prepared"):
                # Create a list of actual spell objects
                _spells = []
                for spell_name in val:
                    msg = 'Spell "{}" not defined. Please add it to ``spells.py``'
                    ThisSpell = self._resolve_mechanic(
                        mechanic=spell_name,
                        SuperClass=spells.Spell,
                        warning_message=msg,
                    )
                    _spells.append(ThisSpell)
                # Sort by name
                _spells.sort(key=lambda spell: spell.name)
                # Save list of spells to character atribute
                if attr == "spells":
                    # Instantiate them all for the spells list
                    self._spells = tuple(S() for S in _spells)
                else:
                    # Instantiate them all for the spells list
                    self._spells_prepared = tuple(S() for S in _spells)
            elif attr == "infusions":
                if hasattr(self, "Artificer"):
                    _infusions = []
                    for infusion_name in val:
                        msg = (
                            "Infusion '{}' not defined. Please add it to"
                            " ``infusions.py``"
                        )
                        ThisInfusion = self._resolve_mechanic(
                            mechanic=infusion_name,
                            SuperClass=infusions.Infusion,
                            warning_message=msg,
                        )
                        _infusions.append(ThisInfusion)
                    _infusions.sort(key=lambda infusion: infusion.name)
                    self.infusions = tuple(i() for i in _infusions)
            elif type(val) not in (type, ModuleType):
                # Some other generic attribute
                is_unknown = not hasattr(self, attr) and not attr.startswith("_")
                if is_unknown:
                    warnings.warn(
                        f"Setting unknown character attribute {attr}", RuntimeWarning
                    )
                # Lookup general attributes
                setattr(self, attr, val)

    def spell_save_dc(self, class_type):
        ability_mod = getattr(self, class_type.spellcasting_ability).modifier
        return 8 + self.proficiency_bonus + ability_mod

    def spell_attack_bonus(self, class_type):
        ability_mod = getattr(self, class_type.spellcasting_ability).modifier
        return self.proficiency_bonus + ability_mod

    def is_proficient(self, weapon: Weapon):
        """Is the character proficient with this item?

        Considers class proficiencies and race proficiencies.

        Parameters
        ----------
        weapon
          The weapon to be tested for proficiency.

        Returns
        -------
        Boolean: is this character proficient with this weapon?

        """
        all_proficiencies = self.weapon_proficiencies
        is_proficient = any((isinstance(weapon, W) for W in all_proficiencies))
        return is_proficient

    @property
    def proficiencies_text(self):
        final_text = ""
        all_proficiencies = tuple(self._proficiencies_text)
        if self.has_class:
            all_proficiencies += tuple(self.primary_class._proficiencies_text)
        if self.num_classes > 1:
            for c in self.class_list[1:]:
                all_proficiencies += tuple(c._multiclass_proficiencies_text)
        if self.race is not None:
            all_proficiencies += tuple(self.race.proficiencies_text)
        if self.background is not None:
            all_proficiencies += tuple(self.background.proficiencies_text)
        # Create a single string out of all the proficiencies
        for txt in all_proficiencies:
            if not final_text:
                # Capitalize the first entry
                txt = txt.capitalize()
            else:
                # Put a comma first
                txt = ", " + txt
                # Add this item to the list text
            final_text += txt
        # Add a period at the end
        final_text += "."
        return final_text

    @proficiencies_text.setter
    def proficiencies_text(self, val):
        try:
            profs = val.split(",")
        except AttributeError:
            profs = val
        self._proficiencies_text = profs

    @property
    def features_text(self):
        s = "\n\n--".join(
            [f.name + ("**" if f.needs_implementation else "") for f in self.features]
        )
        if s != "":
            s = "(See Features Page)\n\n--" + s
            s += "\n\n=================\n\n"
        return s

    @property
    def magic_items_text(self):
        s = ", ".join(
            [f.name for f in sorted(self.magic_items, key=(lambda x: x.name))]
        )
        if s:
            s += ", "
        return s

    def wear_armor(self, new_armor):
        """Accepts a string or Armor class and replaces the current armor.

        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Armor` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Armor` can be provided
        directly.

        """
        if new_armor not in ("", "None", None):
            if isinstance(new_armor, armor.Armor):
                new_armor = new_armor
            else:
                msg = 'Unnown armor "{}". Please add it to ``armor.py``.'
                NewArmor = self._resolve_mechanic(
                    mechanic=new_armor,
                    SuperClass=armor.Armor,
                    warning_message=msg,
                )
                new_armor = NewArmor()
            self.armor = new_armor

    def wield_shield(self, shield):
        """Accepts a string or Shield class and replaces the current armor.

        If a string is given, then a subclass of
        :py:class:`~dungeonsheets.armor.Shield` is retrived from the
        ``armor.py`` file. Otherwise, an subclass of
        :py:class:`~dungeonsheets.armor.Shield` can be provided
        directly.

        """
        if shield not in ("", "None", None):
            msg = 'Unknown shield "{}". Please ad it to ``shields.py``.'
            NewShield = self._resolve_mechanic(shield, SuperClass=armor.Shield, warning_message=msg)
            self.shield = NewShield()

    def wield_weapon(self, weapon):
        """Accepts a string and adds it to the list of wielded weapons.

        Parameters
        ----------
        weapon : str
          Case-insensitive string with a name of the weapon.

        """
        # Retrieve the weapon class from the weapons module
        msg = 'Unknown weapon "{}". Please add it to ``weapons.py``.'
        ThisWeapon = self._resolve_mechanic(
            mechanic=weapon,
            SuperClass=weapons.Weapon,
            warning_message=msg,
        )
        # Save it to the array
        self._weapons.append(ThisWeapon(wielder=self))

    @property
    def weapons(self):
        my_weapons = self._weapons.copy()
        # Account for unarmed strike
        if len(my_weapons) == 0 or hasattr(self, "Monk"):
            my_weapons.append(weapons.Unarmed(wielder=self))
        return my_weapons
    
    @property
    def hit_dice(self):
        """What type and how many dice to use for re-gaining hit points.
        
        To change, set hit_dice_num and hit_dice_faces."""
        dice_s = " + ".join([f"{c.level}d{c.hit_dice_faces}" for c in self.class_list])
        dice_s = combine_dice(dice_s)
        return dice_s

    @property
    def hit_dice_faces(self):
        # Not a valid function if multiclass
        if self.num_classes > 1:
            warnings.warn("hit_dice_faces is not valid for multiclass characters")
        return self.primary_class.hit_dice_faces

    @hit_dice_faces.setter
    def hit_dice_faces(self, faces):
        self.primary_class.hit_dice_faces = faces

    @property
    def proficiency_bonus(self):
        if self.level < 5:
            prof = 2
        elif 5 <= self.level < 9:
            prof = 3
        elif 9 <= self.level < 13:
            prof = 4
        elif 13 <= self.level < 17:
            prof = 5
        elif 17 <= self.level:
            prof = 6
        return prof

    def can_assume_shape(self, shape: monsters.Monster):
        return hasattr(self, "Druid") and self.Druid.can_assume_shape(shape)

    @property
    def all_wild_shapes(self):
        if hasattr(self, "Druid"):
            return self.Druid.all_wild_shapes
        else:
            return ()

    @property
    def wild_shapes(self):
        if hasattr(self, "Druid"):
            return self.Druid.wild_shapes
        else:
            return ()

    @wild_shapes.setter
    def wild_shapes(self, new_shapes):
        if hasattr(self, "Druid"):
            self.Druid.wild_shapes = new_shapes

    @property
    def infusions_text(self):
        if hasattr(self, "Artificer"):
            return tuple([i.name for i in self.infusions])
        else:
            return ()

    @classmethod
    def load(Cls, char_props: MutableMapping):
        """Factory Creates a character from the character definition.

        Parameters
        ==========
        char_props
          Keys and values holding all the attributes of the
          character. E.g. ``char_props['strength'] = 16``

        Returns
        =======
        char
          The initialized ``Character`` object with associated
          parameters.

        """
        # Parse the sheet type
        char_props.pop("sheet_type", "")
        # Load classes
        classes = char_props.get("classes", [])
        # backwards compatability
        if (len(classes) == 0) and ("character_class" in char_props):
            char_props["classes"] = [
                char_props.pop("character_class").lower().capitalize()
            ]
            char_props["levels"] = [str(char_props.pop("level"))]
        # Create the character with loaded properties
        char = Cls(**char_props)
        return char

    def save(self, filename, template_file="character_template.txt"):
        # Create the template context
        context = dict(
            char=self,
        )
        # Render the template
        src_path = os.path.join(os.path.dirname(__file__), "forms/")
        text = (
            jinja2.Environment(loader=jinja2.FileSystemLoader(src_path or "./"))
            .get_template(template_file)
            .render(context)
        )
        # Save the file
        with open(filename, mode="w") as f:
            f.write(text)

    def to_pdf(self, filename, **kwargs):
        from dungeonsheets.make_sheets import make_sheet

        if filename.endswith(".pdf"):
            filename = filename.replace("pdf", "py")
        make_sheet(filename, character=self, flatten=kwargs.get("flatten", True))


# Add backwards compatibility for tests
class Artificer(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Artificer"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Barbarian(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Barbarian"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Bard(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Bard"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Cleric(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Cleric"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Druid(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Druid"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Fighter(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Fighter"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Monk(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Monk"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Paladin(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Paladin"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Ranger(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Ranger"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Rogue(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Rogue"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Sorceror(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Sorceror"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Warlock(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Warlock"]
        attrs["levels"] = [level]
        super().__init__(**attrs)


class Wizard(Character):
    def __init__(self, level=1, **attrs):
        attrs["classes"] = ["Wizard"]
        attrs["levels"] = [level]
        super().__init__(**attrs)



from dungeonsheets.spells.spells import Spell


class PassWithoutTrace(Spell):
    """A veil of shadows and silence radiates from you, masking you and your companions
    from detection.
    For the duration, each creature you choose within 30 feet of
    you (including you) has a +10 bonus to Dexterity (Stealth) checks and can't be
    tracked except by magical means. A creature that receives this bonus leaves
    behind no tracks or other traces of its passage.
    """

    name = "Pass Without Trace"
    level = 2
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = "Ashes from a burned leaf of mistletoe and a sprig of spruce"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Druid", "Ranger")


class Passwall(Spell):
    """A passage appears at a point of your choice that you can see on a wooden,
    plaster, or stone surface (such as a wall, a ceiling, or a floor) within range,
    and lasts for the duration. You choose the opening's dimensions: up to 5 feet
    wide, 8 feet tall, and 20 feet deep. The passage creates no instability in a
    structure surrounding it.

    When the opening disappears, any creatures or objects
    still in the passage created by the spell are safely ejected to an unoccupied
    space nearest to the surface on which you cast the spell.
    """

    name = "Passwall"
    level = 5
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A pinch of sesame seeds"
    duration = "1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Wizard",)


class PhantasmalForce(Spell):
    """You craft an illusion that takes root in the mind of a creature
    that you can see within range. The target must make an
    Intelligence saving throw. On a failed save, you create a
    phantasmal object, creature, or other visible phenomenon of your
    choice that is no larger than a 10-foot cube and that is
    perceivable only to the target for the duration. This spell has no
    effect on undead or constructs.

    The phantasm includes sound, temperature, and other stimuli, also
    evident only to the creature.

    The target can use its action to examine the phantasm with an
    Intelligence (Investigation) check against your spell save DC. If
    the check succeeds, the target realizes that the phantasm is an
    illusion, and the spell ends.

    While a target is affected by the spell, the target treats the
    phantasm as if it were real. The target rationalizes any illogical
    outcomes from interacting with the phantasm. For example, a target
    attempting to walk across a phantasmal bridge that spans a chasm
    falls once it steps onto the bridge. If the target survives the
    fall, it still believes that the bridge exists and comes up with
    some other explanation for its fall - it was pushed, it slipped,
    or a strong wind might have knocked it off.

    An affected target is so convinced of the phantasm's reality that
    it can even take damage from the illusion. A phantasm created to
    appear as a creature can attack the target. Similarly, a phantasm
    created to appear as fire, a pool of acid, or lava can burn the
    target. Each round on your turn, the phantasm can deal 1d6 psychic
    damage to the target if it is in the phantasm's area or within 5
    feet of the phantasm, provided that the illusion is of a creature
    or hazard that could logically deal damage, such as by
    attacking. The target perceives the damage as a type appropriate
    to the illusion.

    """

    name = "Phantasmal Force"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A bit of fleece"
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Sorcerer", "Wizard")


class PhantasmalKiller(Spell):
    """You tap into the nightmares of a creature you can see within range
    and create an illusory manifestation of its deepest fears, visible
    only to that creature.  The target must make a Wisdom saving
    throw. On a failed save, the target becomes frightened for the
    duration. At the end of each of the target's turns before the
    spell ends, the target must succeed on a Wisdom saving throw or
    take 4d10 psychic damage. On a successful save, the spell ends.

    **At Higher Levels:** When you cast this spell using a spell slot
    of 5th level or higher, the damage increases by 1d10 for each slot
    level above 4th.

    """

    name = "Phantasmal Killer"
    level = 4
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ("Wizard",)


class PhantomSteed(Spell):
    """A Large quasi-real, horselike creature appears on the ground in an unoccupied
    space of your choice within range. You decide the creature's appearance, but it
    is equipped with a saddle, bit, and bridle. Any of the equipment created by the
    spell vanishes in a puff of smoke if it is carried more than 10 feet away from
    the steed.

    For the duration, you or a creature you choose can ride the steed.
    The creature uses the statistics for a riding horse, except it has a speed of
    100 feet and can travel 10 miles in an hour, or 13 miles at a fast pace. When
    the spell ends, the steed gradually fades, giving the rider 1 minute to
    dismount. The spell ends if you use an action to dismiss it or if the steed
    takes any damage.
    """

    name = "Phantom Steed"
    level = 3
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = True
    magic_school = "Illusion"
    classes = ("Wizard",)


class PlanarAlly(Spell):
    """You beseech an otherworldly entity for aid.
    The being must be known to you: a
    god, a primordial, a demon prince, or some other being of cosmic power. That
    entity sends a celestial, an elemental, or a fiend loyal to it to aid you,
    making the creature appear in an unoccupied space within range. If you know a
    specific creature's name, you can speak that name when you cast this spell to
    request that creature, though you might get a different creature anyway (DM's
    choice).

    When the creature appears, it is under no compulsion to behave in any
    particular way. You can ask the creature to perform a service in exchange for
    payment, but it isn't obliged to do so. The requested task could range from
    simple (fly us across the chasm, or help us fight a battle) to complex (spy on
    our enemies, or protect us during our foray into the dungeon). You must be able
    to communicate with the creature to bargain for its services.

    Payment can take
    a variety of forms. A celestial might require a sizable donation of gold or
    magic items to an allied temple, while a fiend might demand a living sacrifice
    or a gift of treasure. Some creatures might exchange their service for a quest
    undertaken by you.

    As a rule of thumb, a task that can be measured in minutes
    requires a payment worth 100 gp per minute. A task measured in hours requires
    1,000 gp per hour. And a task m easured in days (up to 10 days) requires 10,000
    gp per day. The DM can adjust these payments based on the circumstances under
    which you cast the spell. If the task is aligned with the creature's ethos, the
    payment might be halved or even waived. Nonhazardous tasks typically require
    only half the suggested payment, while especially dangerous tasks might require
    a greater gift. Creatures rarely accept tasks that seem suicidal.

    After the
    creature completes the task, or when the agreed-upon duration of service
    expires, the creature returns to its home plane after reporting back to you, if
    appropriate to the task and if possible. If you are unable to agree on a price
    for the creature's service, the creature immediately returns to its home plane.


    A creature enlisted to join your group counts as a member of it, receiving a
    full share of experience points awarded.
    """

    name = "Planar Ally"
    level = 6
    casting_time = "10 minutes"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric",)


class PlanarBinding(Spell):
    """With this spell, you attempt to bind a celestial, an elemental, a fey, or a
    fiend to your service.
    The creature must be within range for the entire casting
    of the spell. (Typically, the creature is first summoned into the center of an
    inverted magic circle in order to keep it trapped while this spell is cast.) At
    the completion of the casting, the target must make a Charisma saving throw. On
    a failed save, it is bound to serve you for the duration. If the creature w as
    summoned or created by another spell, that spell's duration is extended to match
    the duration of this spell.

    A bound creature must follow your instructions to
    the best of its ability. You might command the creature to accompany you on an
    adventure, to guard a location, or to deliver a message. The creature obeys the
    letter of your instructions, but if the creature is hostile to you, it strives
    to twist your words to achieve its own objectives. If the creature carries out
    your instructions completely before the spell ends, it travels to you to report
    this fact if you are on the same plane of existence. If you are on a different
    plane of existence, it returns to the place where you bound it and remains there
    until the spell ends.

    **At Higher Levels:** When you cast this spell using a spell
    slot of a higher level, the duration increases to:
    10 days with a 6th-level
    slot,
    30 days with a 7th-level slot,
    180 days with an 8th-level slot,
    1 year
    and 1 day with a 9th-level spell slot.
    """

    name = "Planar Binding"
    level = 5
    casting_time = "1 hour"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A jewel worth at least 1,000 gp, which the spell consumes"
    duration = "24 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Bard", "Cleric", "Druid", "Wizard")


class PlaneShift(Spell):
    """You and up to eight willing creatures who link hands in a circle are transported
    to a different plane of existence. You can specify a target destination in
    general terms, such as the City of Brass on the Elemental Plane of Fire or the
    palace of Dispater on the second level of the Nine Hells, and you appear in or
    near that destination. If you are trying to reac the City of Brass, for example,
    you might arrive in its Street of Steel, before its Gate of Ashes, or looking
    at the city from across the Sea of Fire, at the DM's discretion.

    Alternatively,
    if you know the sigil sequence of a teleportation circle on another plane of
    existence, this spell can take you to that circle. If the teleportation circle
    is too small to hold all the creatures you transported, they appear in the
    closest unoccupied spaces next to the circle.

    You can use this spell to banish
    an unwilling creature to another plane. Choose a creature within your reach and
    make a melee spell attack against it. On a hit, the creature must make a
    Charisma saving throw. If the creature fails the save, it is transported to a
    random location on the plane of existence you specify. A creature so transported
    must find its own way back to your current plane of existence.
    """

    name = "Plane Shift"
    level = 7
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "A forked, metal rod worth at least 250 gp, attuned to a particular plane of"
        " existence"
    )
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric", "Druid", "Sorcerer", "Warlock", "Wizard")


class PlantGrowth(Spell):
    """This spell channels vitality into plants within a specific area. There are two
    possible uses for the spell, granting either immediate or long-term benefits.


    If you cast this spell using 1 action, choose a point within range. All normal
    plants in a 100-foot radius centered on that point become thick and overgrown. A
    creature moving through the area must spend 4 feet of movement for every 1 foot
    it moves.

    You can exclude one or more areas of any size within the spell's
    area from being affected.

    If you cast this spell over 8 hours, you enrich the
    land. All plants in a half-mile radius centered on a point within range become
    enriched for 1 year. The plants yield twice the normal amount of food when
    harvested.
    """

    name = "Plant Growth"
    level = 3
    casting_time = "Special"
    casting_range = "150 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Druid", "Ranger")


class PoisonSpray(Spell):
    """You extend your hand toward a creature you can see within range and project a
    puff of noxious gas from your palm. The creature must succeed on a Constitution
    saving throw or take 1d12 poison damage.

    **At Higher Levels:** This spell's damage
    increases by 1d12 when you reach 5th level (2d12), 11th level (3d12), 17th level
    (4d12).
    """

    name = "Poison Spray"
    level = 0
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class Polymorph(Spell):
    """This spell transforms a creature that you can see within range into a new form.
    An unwilling creature must make a Wisdom saving throw to avoid the effect. A
    shapechanger automatically succeeds on this saving throw.

    The transformation
    lasts for the duration, or until the target drops to 0 hit points or dies. The
    new form can be any beast whose challenge rating is equal to or less than the
    target's (or the target's level, if it doesn't have a challenge rating). The
    target's game statistics, including mental ability scores, are replaced by the
    statistics of the chosen beast. It retains its alignment and personality.

    The
    target assumes the hit points of its new form. When it reverts to its normal
    form, the creature returns to the number of hit points it had before it
    transformed. If it reverts as a result of dropping to 0 hit points, any excess
    damage carries over to its normal form. As long as the excess damage doesn't
    reduce the creature's normal form to 0 hit points, it isn't knocked unconscious.


    The creature is limited in the actions it can perform by the nature of its new
    form, and it can't speak, cast spells, or take any other action that requires
    hands or speech.

    The target's gear melds into the new form. The creature can't
    activate, use, wield, or otherwise benefit from any of its equipment. This spell
    can't affect a target that has 0 hit points.
    """

    name = "Polymorph"
    level = 4
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S", "M")
    materials = "A caterpillar cocoon"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Druid", "Sorcerer", "Wizard")


class PowerWordHeal(Spell):
    """A wave of healing energy washes over the creature you touch. The target regains
    all its hit points. If the creature is charmed, frightened, paralyzed, or
    stunned, the condition ends. If the creature is prone, it can use its reaction
    to stand up. This spell has no effect on undead or constructs.
    """

    name = "Power Word Heal"
    level = 9
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Bard",)


class PowerWordKill(Spell):
    """You utter a word of power that can compel one creature you can see within range
    to die instantly. If the creature you chose has 100 hit points or fewer, it
    dies. Otherwise, the spell has no effect.
    """

    name = "Power Word Kill"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class PowerWordPain(Spell):
    """You speak a word of power that causes waves of intense pain to assail one
    creature you can see within range. If the target has 100 hit points or fewer, it
    is subject to crippling pain. Otherwise, the spell has no effect on it. A
    target is also unaffected if it is immune to being charmed.
    While the target is
    affected by crippling pain, any speed it has can be no higher than 10 feet. The
    target also has disadvantage on attack rolls, ability checks, and saving throws,
    other than Constitution saving throws. Finally, if the target tries to cast a
    spell, it must first succeed on a Constitution saving throw, or the casting
    fails and the spell is wasted.
    A target suffering this pain can make a
    Constitution saving throw at the end of each of its turns. On a successful save,
    the pain ends.
    """

    name = "Power Word Pain"
    level = 7
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Sorcerer", "Warlock", "Wizard")


class PowerWordStun(Spell):
    """You speak a word of power that can overwhelm the mind of one creature you can
    see within range, leaving it dumbfounded. If the target has 150 hit points or
    fewer, it is stunned. Otherwise, the spell has no effect. The stunned target
    must make a Constitution saving throw at the end of each of its turns. On a
    successful save, this stunning effect ends.
    """

    name = "Power Word Stun"
    level = 8
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class PrayerOfHealing(Spell):
    """Up to six creatures of your choice that you can see within range each regain hit
    points equal to 2d8 + your spellcasting ability modifier. This spell has no
    effect on undead or constructs.

    **At Higher Levels:** When you cast this spell
    using a spell slot of 3rd level or higher, the healing increases by 1d8 for each
    slot level above 2nd.
    """

    name = "Prayer Of Healing"
    level = 2
    casting_time = "10 minutes"
    casting_range = "30 feet"
    components = ("V",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Cleric",)


class Prestidigitation(Spell):
    """This spell is a minor magical trick that novice spellcasters use for practice.
    You create one of the following magical effects within range:
     -You create an
    instantaneous, harmless sensory effect, such as a shower of sparks, a puff of
    wind, faint musical notes, or an odd odor.
    -You instantaneously light or snuff
    out a candle, a torch, or a small campfire.
    -You instantaneously clean or soil
    an object no larger than 1 cubic foot.
    -You chill, warm, or flavor up to 1
    cubic foot of nonliving material for 1 hour.
    -You make a color, a small mark,
    or a symbol appear on an object or a surface for 1 hour.
    -You create a
    nonmagical trinket or an illusory image that can fit in your hand and that lasts
    until the end of your next turn.
    If you cast this spell multiple times, you
    can have up to three of its non-instantaneous effects active at a time, and you
    can dismiss such an effect as an action.
    """

    name = "Prestidigitation"
    level = 0
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ("V", "S")
    materials = ""
    duration = "Up to 1 hour"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class PrimalSavagery(Spell):
    """You channel primal magic to cause your teeth or fingernails to sharpen, ready to
    deliver a corrosive attack. Make a melee spell attack against one creature
    within 5 feet of you. On a hit, the target takes 1d10 acid damage. After you
    make the attack, your teeth or fingernails return to normal. The spell's damage
    increases by 1d10 when you reach 5th level (2d10), 11th level (3d10), and 17th
    level (4d10).
    """

    name = "Primal Savagery"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ("S",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid",)


class PrimordialWard(Spell):
    """You have resistance to acid, cold, fire, lightning, and thunder damage for the
    spell's duration.
    When you take damage of one of those types, you can use your
    reaction to gain immunity to that type
    of damage, including against the
    triggering damage. If you do so, the resistances end, and you have the immunity
    until the end of your next turn, at which time the spell ends.
    """

    name = "Primordial Ward"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Druid",)


class PrismaticSpray(Spell):
    """Eight multicolored rays of light flash from your hand. Each ray is a different
    color and has a different power and purpose. Each creature in a 60-foot cone
    must make a Dexterity saving throw. For each target, roll a d8 to determine
    which color ray affects it.

    1. Red. The target takes 10d6 fire damage on a
    failed save, or half as much damage on a successful one.

    2. Orange. The target
    takes 10d6 acid damage on a failed save, or half as much damage on a successful
    one.

    3. Yellow. The target takes 10d6 lightning damage on a failed save, or
    half as much damage on a successful one.

    4. Green. The target takes 10d6 poison
    damage on a failed save, or half as much damage on a successful one.

    5. Blue.
    The target takes 10d6 cold damage on a failed save, or half as much damage on a
    successful one.

    6. Indigo. On a failed save, the target is restrained. It must
    then make a Constitution saving throw at the end of each of its turns. If it
    successfully saves three times, the spell ends. If it fails its save three
    times, it permanently turns to stone and is subjected to the petrified
    condition. The successes and failures don't need to be consecutive; keep track
    of both until the target collects three of a kind.

    7. Violet. On a failed save,
    the target is blinded. It must then make a Wisdom saving throw at the start of
    your next turn. A successful save ends the blindness. If it fails that save, the
    creature is transported to another plane of existence of the DM's choosing and
    is no longer blinded. (Typically, a creature that is on a plane that isn't its
    home plane is banished home, while other creatures are usually cast into the
    Astral or Ethereal planes.)

    8. Special. The target is struck by two rays. Roll
    twice more, rerolling any 8.
    """

    name = "Prismatic Spray"
    level = 7
    casting_time = "1 action"
    casting_range = "Self (60-foot cone)"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class PrismaticWall(Spell):
    """A shimmering, multicolored plane of light forms a vertical opaque
    wall-up to 90 feet long, 30 feet high, and 1 inch thick-centered
    on a point you can see within range. Alternatively, you can shape
    the wall into a sphere up to 30 feet in diameter centered on a
    point you choose within range. The wall remains in place for the
    duration. If you position the wall so that it passes through a
    space occupied by a creature, the spell fails, and your action and
    the spell slot are wasted.

    The wall sheds bright light out to a range of 100 feet and dim
    light for an additional 100 feet. You and creatures you designate
    at the time you cast the spell can pass through and remain near
    the wall without harm. If another creature that can see the wall
    moves to within 20 feet of it or starts its turn there, the
    creature must succeed on a Constitution saving throw or become
    blinded for 1 minute.

    The wall consists of seven layers, each with a different
    color. When a creature attempts to reach into or pass through the
    wall, it does so one layer at a time through all the wall's
    layers. As it passes or reaches through each layer, the creature
    must make a Dexterity saving throw or be affected by that layer's
    properties as described below.

    The wall can be destroyed, also one layer at a time, in order from
    red to violet, by means specific to each layer. Once a layer is
    destroyed, it remains so for the duration of the spell. A rod of
    cancellation destroys a prismatic wall, but an antimagic field has
    no effect on it.

    1. Red. The creature takes 10d6 fire damage on a failed save, or
    half as much damage on a successful one. While this layer is in
    place, nonmagical ranged attacks can't pass through the wall. The
    layer can be destroyed by dealing at least 25 cold damage to it.

    2. Orange. The creature takes 10d6 acid damage on a failed save,
    or half as much damage on a successful one. While this layer is in
    place, magical ranged attacks can't pass through the wall. The
    layer is destroyed by a strong wind.

    3. Yellow. The creature takes 10d6 lightning damage on a failed
    save, or half as much damage on a successful one. This layer can
    be destroyed by dealing at least 60 force damage to it.

    4. Green. The creature takes 10d6 poison damage on a failed save,
    or half as much damage on a successful one. A passwall spell, or
    another spell of equal or greater level that can open a portal on
    a solid surface, destroys this layer.

    5. Blue. The creature takes 10d6 cold damage on a failed save, or
    half as much damage on a successful one. This layer can be
    destroyed by dealing at least 25 fire damage to it.

    6. Indigo. On a failed save, the creature is restrained. It must
    then make a Constitution saving throw at the end of each of its
    turns. If it successfully saves three times, the spell ends. If it
    fails its save three times, it permanently turns to stone and is
    subjected to the petrified condition. The successes and failures
    don't need to be consecutive; keep track of both until the
    creature collects three of a kind. While this layer is in place,
    spells can't be cast through the wall. The layer is destroyed by
    bright light shed by a daylight spell or a similar spell of equal
    or higher level.

    7. Violet. On a failed save, the creature is blinded. It must then
    make a Wisdom saving throw at the start of your next turn. A
    successful save ends the blindness. If it fails that save, the
    creature is transported to another plane of the DM's choosing and
    is no longer blinded. (Typically, a creature that is on a plane
    that isn't its home plane is banished home, while other creatures
    are usually cast into the Astral or Ethereal planes.) This layer
    is destroyed by a dispel magic spell or similar spell of equal or
    higher level that can end spells and magical effects.

    """

    name = "Prismatic Wall"
    level = 9
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Wizard",)


class ProduceFlame(Spell):
    """A flickering flame appears in your hand.
    The flame remains there for the
    duration and harms neither you nor your equipment. The flame sheds bright light
    in a 10-foot radius and dim light for an additional 10 feet. The spell ends if
    you dismiss it as an action or if you cast it again.

    You can also attack with
    the flame, although doing so ends the spell. When you cast this spell, or as an
    action on a later turn, you can hurl the flame at a creature within 30 feet of
    you. Make a ranged spell attack. On a hit, the target takes 1d8 fire damage.

    At
    Higher Levels: This spell's damage increases by 1d8 when you reach 5th level
    (2d8), 11th level (3d8), and 17th level (4d8).
    """

    name = "Produce Flame"
    level = 0
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid",)


class ProgrammedIllusion(Spell):
    """You create an illusion of an object, a creature, or some other visible
    phenomenon within range that activates when a specific condition occurs. The
    illusion is imperceptible until then. It must be no larger than a 30-foot cube,
    and you decide when you cast the spell how the illusion behaves and what sounds
    it makes. This scripted performance can last up to 5 minutes.

    When the
    condition you specify occurs, the illusion springs into existence and performs
    in the manner you described. Once the illusion finishes performing, it
    disappears and remains dormant for 10 minutes. After this time, the illusion can
    be activated again.

    The triggering condition can be as general or as detailed
    as you like, though it must be based on visual or audible conditions that occur
    within 30 feet of the area. For example, you could create an illusion of
    yourself to appear and warn off others who attempt to open a trapped door, or
    you could set the illusion to trigger only when a creature says the correct word
    or phrase.

    Physical interaction with the image reveals it to be an illusion,
    because things can pass through it. A creature that uses its action to examine
    the image can determine that it is an illusion with a successful Intelligence
    (Investigation) check against your spell save DC. If a creature discerns the
    illusion for what it is, the creature can see through the image, and any noise
    it makes sounds hollow to the creature.
    """

    name = "Programmed Illusion"
    level = 6
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("V", "S", "M")
    materials = "A bit of fleece and jade dust worth at least 25 gp"
    duration = "Until dispelled"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Wizard")


class ProjectImage(Spell):
    """You create an illusory copy of yourself that lasts for the duration.
    The copy
    can appear at any location within range that you have seen before, regardless of
    intervening obstacles. The illusion looks and sounds like you but is
    intangible. If the illusion takes any damage, it disappears, and the spell ends.


    You can use your action to move this illusion up to twice your speed, and make
    it gesture, speak, and behave in whatever way you choose. It mimics your
    mannerisms perfectly.

    You can see through its eyes and hear through its ears as
    if you were in its space. On your turn as a bonus action, you can switch from
    using its senses to using your own, or back again. While you are using its
    senses, you are blinded and deafened in regard to your own surroundings.


    Physical interaction with the image reveals it to be an illusion, because things
    can pass through it. A creature that uses its action to examine the image can
    determine that it is an illusion with a successful Intelligence (Investigation)
    check against your spell save DC. If a creature discerns the illusion for what
    it is, the creature can see through the image, and any noise it makes sounds
    hollow to the creature.
    """

    name = "Project Image"
    level = 7
    casting_time = "1 action"
    casting_range = "500 miles"
    components = ("V", "S", "M")
    materials = "A small replica of you made from materials worth at least 5 gp"
    duration = "Concentration, up to 1 day"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Wizard")


class ProtectionFromEnergy(Spell):
    """For the duration, the willing creature you touch has resistance to one damage
    type of your choice: acid, cold, fire, lightning, or thunder.
    """

    name = "Protection From Energy"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Druid", "Ranger", "Sorcerer", "Wizard")


class ProtectionFromEvilAndGood(Spell):
    """Until the spell ends, one willing creature you touch is protected against
    certain types of creatures: aberrations, celestials, elementals, fey, fiends,
    and undead.

    The protection grants several benefits. Creatures of those types
    have disadvantage on attack rolls against the target. The target also can't be
    charmed, frightened, or possessed by them. If the target is already charmed,
    frightened, or possessed by such a creature, the target has advantage on any new
    saving throw against the relevant effect.
    """

    name = "Protection From Evil And Good"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "Holy water or powdered silver and iron, which the spell consumes"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Paladin", "Warlock", "Wizard")


class ProtectionFromPoison(Spell):
    """You touch a creature. If it is poisoned, you neutralize the poison. If more than
    one poison afflicts the target, you neutralize one poison that you know is
    present, or you neutralize one at random.

    For the duration, the target has
    advantage on saving throws against being poisoned, and it has resistance to
    poison damage.
    """

    name = "Protection From Poison"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "1 hour"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Cleric", "Druid", "Paladin", "Ranger")


class PsychicScream(Spell):
    """You unleash the power of your mind to blast the intellect of up to ten creatures
    of your choice that you can see within range. Creatures that have an
    Intelligence score of 2 or lower are unaffected.
    Each target must make an
    Intelligence saving throw. On a failed save, a target takes 14d6 psychic damage
    and is stunned. On a successful save, a target takes half as much damage and
    isn't stunned. If a target is killed by this damage, its head explodes, assuming
    it has one.
    A stunned target can make an Intelligence saving throw at the end
    of each of its turns. On a successful save, the stunning effect ends.
    """

    name = "Psychic Scream"
    level = 9
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("S",)
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Enchantment"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class PurifyFoodAndDrink(Spell):
    """All nonmagical food and drink within a 5-foot-radius sphere centered on a point
    of your choice within range is purified and rendered free of poison and disease.
    """

    name = "Purify Food And Drink"
    level = 1
    casting_time = "1 action"
    casting_range = "10 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = True
    magic_school = "Transmutation"
    classes = ("Cleric", "Druid", "Paladin")


class Pyrotechnics(Spell):
    """Choose an area of nonmagical flame that you can see and that fits within a
    5-foot cube within range. You can extinguish the fire in that area, and you
    create either fireworks or smoke when you do so.
    Fireworks. The target explodes
    with a dazzling display of colors. Each creature within 10 feet of the target
    must succeed on a Constitution saving throw or become blinded until the end of
    your next turn.
    Smoke. Thick black smoke spreads out from the target in a
    20-foot radius, moving around corners. The area of the smoke is heavily
    obscured. The smoke persists for 1 minute or until a strong wind disperses it.
    """

    name = "Pyrotechnics"
    level = 2
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Bard", "Sorcerer", "Wizard")
